/*
 * Caveworld
 *
 * Copyright (c) 2014 kegare
 * https://github.com/kegare
 *
 * This mod is distributed under the terms of the Minecraft Mod Public License 1.0, or MMPL.
 * Please check the contents of the license located in http://www.mod-buildcraft.com/MMPL-1.0.txt
 */

package com.kegare.caveworld.handler;

import java.security.SecureRandom;
import java.util.Random;
import java.util.logging.Level;

import net.minecraft.block.Block;
import net.minecraft.block.BlockBed;
import net.minecraft.client.Minecraft;
import net.minecraft.client.resources.I18n;
import net.minecraft.entity.Entity;
import net.minecraft.entity.EntityLiving;
import net.minecraft.entity.EntityLivingBase;
import net.minecraft.entity.item.EntityItem;
import net.minecraft.entity.passive.EntityBat;
import net.minecraft.entity.player.EntityPlayer;
import net.minecraft.entity.player.EntityPlayerMP;
import net.minecraft.entity.player.EnumStatus;
import net.minecraft.item.Item;
import net.minecraft.item.ItemPickaxe;
import net.minecraft.item.ItemStack;
import net.minecraft.nbt.NBTTagCompound;
import net.minecraft.util.ChunkCoordinates;
import net.minecraft.util.DamageSource;
import net.minecraft.util.MathHelper;
import net.minecraft.world.World;
import net.minecraft.world.WorldServer;
import net.minecraftforge.client.event.RenderGameOverlayEvent;
import net.minecraftforge.client.event.sound.SoundLoadEvent;
import net.minecraftforge.event.ForgeSubscribe;
import net.minecraftforge.event.entity.EntityJoinWorldEvent;
import net.minecraftforge.event.entity.living.LivingDropsEvent;
import net.minecraftforge.event.entity.living.LivingEvent.LivingUpdateEvent;
import net.minecraftforge.event.entity.player.PlayerEvent.BreakSpeed;
import net.minecraftforge.event.entity.player.PlayerInteractEvent;
import net.minecraftforge.event.entity.player.PlayerInteractEvent.Action;
import net.minecraftforge.event.entity.player.PlayerSleepInBedEvent;
import net.minecraftforge.event.world.BlockEvent.HarvestDropsEvent;
import net.minecraftforge.event.world.WorldEvent;

import com.kegare.caveworld.block.CaveBlock;
import com.kegare.caveworld.core.CaveMiningManager;
import com.kegare.caveworld.core.CaveOreManager;
import com.kegare.caveworld.core.Config;
import com.kegare.caveworld.util.CaveLog;
import com.kegare.caveworld.world.WorldProviderCaveworld;

import cpw.mods.fml.client.FMLClientHandler;
import cpw.mods.fml.relauncher.Side;
import cpw.mods.fml.relauncher.SideOnly;

public class CaveEventHooks
{
	@SideOnly(Side.CLIENT)
	@ForgeSubscribe
	public void onSoundLoad(SoundLoadEvent event)
	{
		try
		{
			event.manager.addSound("caveworld:caveworld_portal.ogg");
			event.manager.addSound("caveworld:ambient/cave.ogg");
		}
		catch (Exception e)
		{
			CaveLog.log(Level.WARNING, e, "An error occurred trying to loading cave sounds");
		}
	}

	@SideOnly(Side.CLIENT)
	@ForgeSubscribe
	public void onRenderGameOverlay(RenderGameOverlayEvent.Text event)
	{
		Minecraft mc = FMLClientHandler.instance().getClient();

		if (mc.thePlayer != null && mc.thePlayer.dimension == Config.dimensionCaveworld)
		{
			ItemStack current = mc.thePlayer.getCurrentEquippedItem();

			if (mc.gameSettings.showDebugInfo)
			{
				event.left.add("dim: Caveworld");
			}
			else if ((mc.gameSettings.advancedItemTooltips || mc.thePlayer.isSneaking()) && current != null && current.getItem() instanceof ItemPickaxe)
			{
				int count = CaveMiningManager.getMiningCount(mc.thePlayer);

				if (count > 0)
				{
					int level = CaveMiningManager.getMiningLevel(mc.thePlayer);
					StringBuilder builder = new StringBuilder();

					builder.append(I18n.getString("caveworld.mining.count")).append(": ").append(count);

					if (level > 0)
					{
						builder.append(" (").append(level).append(')');
					}

					event.right.add(builder.toString());
				}
			}
		}
	}

	@ForgeSubscribe
	public void onHarvestDrops(HarvestDropsEvent event)
	{
		if (event.harvester != null && event.harvester instanceof EntityPlayerMP)
		{
			EntityPlayerMP player = (EntityPlayerMP)event.harvester;

			if (player.dimension == Config.dimensionCaveworld)
			{
				ItemStack current = player.getCurrentEquippedItem();
				Block block = event.block;
				int metadata = event.blockMetadata;

				if (current != null && current.getItem() instanceof ItemPickaxe && CaveOreManager.containsOre(block, metadata))
				{
					CaveMiningManager.addMiningCount(player, 1);
				}
			}
		}
	}

	@ForgeSubscribe
	public void onBreakSpeed(BreakSpeed event)
	{
		EntityPlayer player = event.entityPlayer;

		if (player != null && player.dimension == Config.dimensionCaveworld && player.isSneaking())
		{
			ItemStack current = player.getCurrentEquippedItem();
			Block block = event.block;
			int metadata = event.metadata;

			if (current != null && current.getItem() instanceof ItemPickaxe && CaveOreManager.containsOre(block, metadata))
			{
				int level = CaveMiningManager.getMiningLevel(player);

				if (level > 2)
				{
					event.newSpeed *= 2.0F;
				}
				else if (level > 0)
				{
					event.newSpeed *= 1.75F;
				}
			}
		}
	}

	@ForgeSubscribe
	public void onPlayerInteract(PlayerInteractEvent event)
	{
		if (event.entityPlayer instanceof EntityPlayerMP)
		{
			EntityPlayerMP player = (EntityPlayerMP)event.entityPlayer;
			WorldServer world = player.getServerForPlayer();
			int x = event.x;
			int y = event.y;
			int z = event.z;
			int face = event.face;
			ItemStack current = player.getCurrentEquippedItem();

			if (event.action == Action.RIGHT_CLICK_BLOCK)
			{
				if (!player.isSneaking() && current != null && current.itemID == Block.enderChest.blockID)
				{
					if (face == 0)
					{
						--y;
					}
					else if (face == 1)
					{
						++y;
					}
					else if (face == 2)
					{
						--z;
					}
					else if (face == 3)
					{
						++z;
					}
					else if (face == 4)
					{
						--x;
					}
					else if (face == 5)
					{
						++x;
					}

					if (CaveBlock.portalCaveworld.tryToCreatePortal(world, x, y, z))
					{
						world.playSoundEffect(x + 0.5D, y + 0.5D, z + 0.5D, CaveBlock.portalCaveworld.stepSound.getPlaceSound(), 1.0F, 2.0F);

						event.setCanceled(true);
					}
				}
				else if (player.dimension == Config.dimensionCaveworld && world.getBlockId(x, y, z) == Block.bed.blockID)
				{
					int metadata = world.getBlockMetadata(x, y, z);

					if (!BlockBed.isBlockHeadOfBed(metadata))
					{
						int var1 = BlockBed.getDirection(metadata);
						x += BlockBed.footBlockToHeadBlockMap[var1][0];
						z += BlockBed.footBlockToHeadBlockMap[var1][1];

						if (world.getBlockId(x, y, z) != Block.bed.blockID)
						{
							return;
						}

						metadata = world.getBlockMetadata(x, y, z);
					}

					if (BlockBed.isBedOccupied(metadata))
					{
						for (Object obj : world.playerEntities)
						{
							EntityPlayer target = (EntityPlayer)obj;

							if (target.isPlayerSleeping())
							{
								ChunkCoordinates coord = target.playerLocation;

								if (coord.posX == x && coord.posY == y && coord.posZ == z)
								{
									player.addChatMessage("tile.bed.occupied");

									return;
								}
							}
						}

						BlockBed.setBedOccupied(world, x, y, z, false);
					}

					EnumStatus status = player.sleepInBedAt(x, y, z);

					if (status == EnumStatus.OK)
					{
						BlockBed.setBedOccupied(world, x, y, z, true);
					}
					else
					{
						if (status == EnumStatus.NOT_POSSIBLE_NOW)
						{
							player.addChatMessage("tile.bed.noSleep");
						}
						else if (status == EnumStatus.NOT_SAFE)
						{
							player.addChatMessage("tile.bed.notSafe");
						}
					}

					event.setCanceled(true);
				}
			}
		}
	}

	@ForgeSubscribe
	public void onPlayerSleepInBed(PlayerSleepInBedEvent event)
	{
		if (event.entityPlayer instanceof EntityPlayerMP)
		{
			EntityPlayerMP player = (EntityPlayerMP)event.entityPlayer;
			WorldServer world = player.getServerForPlayer();

			if (player.dimension == Config.dimensionCaveworld && !player.capabilities.isCreativeMode)
			{
				NBTTagCompound data = player.getEntityData();
				ChunkCoordinates spawn = player.getBedLocation(player.dimension);

				if (data.getLong("Caveworld:LastSleepTime") + 6000L > world.getTotalWorldTime())
				{
					event.result = EnumStatus.OTHER_PROBLEM;
				}
				else if (spawn != null && player.getDistance(spawn.posX, spawn.posY, spawn.posZ) <= 32.0D)
				{
					event.result = EnumStatus.OTHER_PROBLEM;
				}

				if (event.result == null)
				{
					data.setLong("Caveworld:LastSleepTime", world.getTotalWorldTime());
				}
			}
		}
	}

	@ForgeSubscribe
	public void onEntityJoinWorld(EntityJoinWorldEvent event)
	{
		World world = event.world;
		Entity entity = event.entity;

		if (!world.isRemote && entity.dimension == Config.dimensionCaveworld)
		{
			if (entity instanceof EntityLiving && MathHelper.floor_double(entity.posY) >= world.provider.getActualHeight())
			{
				event.setCanceled(true);
			}
		}
	}

	@ForgeSubscribe
	public void onLivingUpdate(LivingUpdateEvent event)
	{
		if (event.entityLiving.dimension == Config.dimensionCaveworld && event.entityLiving.ticksExisted % 20 == 0)
		{
			if (event.entityLiving instanceof EntityPlayerMP)
			{
				EntityPlayerMP player = (EntityPlayerMP)event.entityLiving;

				if (player.isPlayerSleeping() && player.getSleepTimer() >= 80)
				{
					player.wakeUpPlayer(true, true, false);

					player.setSpawnChunk(player.playerLocation, true, player.dimension);
				}
			}
		}
	}

	@ForgeSubscribe
	public void onLivingDrops(LivingDropsEvent event)
	{
		Random random = new SecureRandom();
		DamageSource source = event.source;
		Entity entity = source.getEntity();
		EntityLivingBase living = event.entityLiving;
		World world = living.worldObj;
		double posX = living.posX;
		double posY = living.posY;
		double posZ = living.posZ;
		int looting = Math.max(event.lootingLevel, 0);

		if (!world.isRemote && living.dimension == Config.dimensionCaveworld && entity != null && entity instanceof EntityPlayerMP)
		{
			if (living instanceof EntityBat)
			{
				event.drops.add(new EntityItem(world, posX, posY + 0.5D, posZ, new ItemStack(Item.coal, random.nextInt(3) + Math.min(looting, 3))));

				if (random.nextInt(3) == Math.min(looting, 2))
				{
					event.drops.add(new EntityItem(world, posX, posY + 0.25D, posZ, new ItemStack(Item.leather, 1)));
				}
			}
		}
	}

	@ForgeSubscribe
	public void onWorldLoad(WorldEvent.Load event)
	{
		World world = event.world;

		if (!world.isRemote && world.provider.dimensionId == Config.dimensionCaveworld)
		{
			CaveMiningManager.loadMiningData();
		}
	}

	@ForgeSubscribe
	public void onWorldUnload(WorldEvent.Unload event)
	{
		World world = event.world;

		if (!world.isRemote && world.provider.dimensionId == Config.dimensionCaveworld)
		{
			CaveMiningManager.saveMiningData();
			CaveMiningManager.clearMiningData();

			WorldProviderCaveworld.writeDimData();
			WorldProviderCaveworld.clearDimData();
		}
	}

	@ForgeSubscribe
	public void onWorldSave(WorldEvent.Save event)
	{
		World world = event.world;

		if (!world.isRemote && world.provider.dimensionId == Config.dimensionCaveworld)
		{
			CaveBlock.portalCaveworld.getInventory().saveInventoryToNBT();
		}
	}
}