/*
 * Caveworld
 *
 * Copyright (c) 2014 kegare
 * https://github.com/kegare
 *
 * This mod is distributed under the terms of the Minecraft Mod Public License 1.0, or MMPL.
 * Please check the contents of the license located in http://www.mod-buildcraft.com/MMPL-1.0.txt
 */

package com.kegare.caveworld.handler;

import java.util.EnumSet;
import java.util.Random;
import java.util.Set;

import net.minecraft.util.WeightedRandom;
import net.minecraft.util.WeightedRandomItem;
import net.minecraft.world.WorldServer;

import com.google.common.base.Strings;
import com.google.common.collect.ImmutableSet;
import com.google.common.collect.Sets;
import com.kegare.caveworld.core.Config;

import cpw.mods.fml.common.IScheduledTickHandler;
import cpw.mods.fml.common.TickType;
import cpw.mods.fml.common.network.PacketDispatcher;

public class CaveMusicTracker implements IScheduledTickHandler
{
	private static final Set<CaveMusic> CAVE_MUSICS = Sets.newHashSet();

	static
	{
		addCaveMusic("ambient.cave", 100);
	}

	private final Random random = new Random();

	@Override
	public void tickStart(EnumSet<TickType> type, Object... tickData) {}

	@Override
	public void tickEnd(EnumSet<TickType> type, Object... tickData)
	{
		if (tickData.length > 0 && tickData[0] instanceof WorldServer)
		{
			onPostWorldTick((WorldServer)tickData[0]);
		}
	}

	private void onPostWorldTick(WorldServer world)
	{
		int dim = world.provider.dimensionId;

		if (dim == Config.dimensionCaveworld && random.nextBoolean())
		{
			String music = getRandomCaveMusic(random);

			if (!Strings.isNullOrEmpty(music))
			{
				PacketDispatcher.sendPacketToAllInDimension(CavePacketHandler.getPacketPlaySound(music), dim);
			}
		}
	}

	@Override
	public EnumSet<TickType> ticks()
	{
		return EnumSet.of(TickType.SERVER, TickType.WORLD);
	}

	@Override
	public String getLabel()
	{
		return "Caveworld Music Tracker";
	}

	@Override
	public int nextTickSpacing()
	{
		return 20000;
	}

	public static boolean addCaveMusic(String name, int weight)
	{
		for (CaveMusic music : CAVE_MUSICS)
		{
			if (music.soundType.equals(name))
			{
				music.itemWeight += weight;

				return false;
			}
		}

		return CAVE_MUSICS.add(new CaveMusic(name, weight));
	}

	public static boolean removeCaveMusic(String name)
	{
		for (CaveMusic music : CAVE_MUSICS)
		{
			if (music.soundType.equals(name))
			{
				return CAVE_MUSICS.remove(music);
			}
		}

		return false;
	}

	public static String getRandomCaveMusic(Random random)
	{
		if (!CAVE_MUSICS.isEmpty())
		{
			return ((CaveMusic)WeightedRandom.getRandomItem(random, CAVE_MUSICS)).soundType;
		}

		return null;
	}

	public static ImmutableSet<CaveMusic> getCaveMusics()
	{
		return new ImmutableSet.Builder().addAll(CAVE_MUSICS).build();
	}

	public static class CaveMusic extends WeightedRandomItem
	{
		public final String soundType;

		public CaveMusic(String name, int weight)
		{
			super(weight);
			this.soundType = name;
		}

		@Override
		public boolean equals(Object target)
		{
			return target instanceof CaveMusic && soundType.equals(((CaveMusic)target).soundType);
		}
	}
}